/* -*- linux-c -*- (Autoset linux indent style when using emacs)	*/

/*
 * Copyright 1999 by Digi International (www.digi.com)
 *    Jeff Randall <Jeff_Randall@digi.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the 
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/************************************************************************
 *
 * /proc/dgdm/{*} functions
 *
 ************************************************************************/
#define __NO_VERSION__
#include "dgdm_driver.h"
#include <linux/proc_fs.h>

/* external variables */
extern int              dgdm_debug;
extern BOARD_T		*dgdm_Board[MAXBOARDS];
extern uchar		dgdm_NumBoards;
extern int		dgdm_major;
extern int		dgdm_tty;
extern int		dgdm_print;
extern int		dgdm_isdn;
extern int		dgdm_control;
extern char		*dgdm_outdevice;
extern char		*dgdm_indevice;
extern char		*dgdm_printdevice;
extern char		*dgdm_specdevice;
extern char		*dgdm_rawdxbdevice;
extern char		*dgdm_ramdevice;
extern char		*dgdm_dspramdevice;
extern char		*dgdm_gdbdevice;
extern char		*dgdm_dmabufdevice;
extern char		*dgdm_regdevice;
extern char		*dgdm_iodevice;
extern char		*dgdm_flashdevice;
extern char		*dgdm_vpdalldevice;

/*
 * OK, here's the deal with /proc files.  We want to use the
 * simple (yet inefficient) interface to /proc where we just get
 * a buffer and fill it in.  Otherwise, we'd have to provide the
 * full set of functions for a /proc driver.
 *
 * When we get called, "buf" points to a single free page which
 * we can write into.  If you are careful, you can write up
 * to PAGE_SIZE (4096) bytes into it.  But to be paranoid,
 * Linux lies and says the "count" is only PROC_BLOK_SIZE (3072).
 *
 * In any case, its easiest to make several /proc/dgdm/ entries
 * if you want to print out more info than can fit in a page.
 *
 * Here's some further information.  We are trying to be compatible
 * with 2.0.xx and 2.1.xx.  There have been improvements in 2.1.x
 * in this area, but it still isn't perfect.  Anyway, our hands
 * are tied by the 2.0.x limitations.  For example, you cannot
 * pass private data to the get_info (2.0.x) routines.  Therefore,
 * you cannot make a generic function to print the data for one
 * board.  If you want to do this, you have to make a wrapper
 * function, one function name per board number.  Yech!
 */

/*
 * Handle some 2.0.x / 2.1.x incompatibilities
 */
#if LINUX_VERSION_CODE >= 0x20100
	#define	PROC_READ_PROTO \
		char *buf, char **start, off_t fpos, int count, \
		int *eof, void *private

	#define	PROC_READ_ARGS \
		buf, start, fpos, count, eof, private
#else
	#define	PROC_READ_PROTO \
		char *buf, char **start, off_t fpos, int count, \
		int dummy

	#define	PROC_READ_ARGS \
		buf, start, fpos, count, dummy

	typedef int (read_proc_t)(PROC_READ_PROTO);
#endif

/*
 *	Make wrappers for per-board entries. Gag me.  Sorry,
 *	2.0.x forced us into this since it has no way to pass
 *	private data to the read_proc (get_info, actually).
 */
#define	PROCWRAP_FUNCTION(FUNC, NUM) \
			static int \
			FUNC ## NUM (PROC_READ_PROTO) \
			{ \
				return FUNC(NUM, PROC_READ_ARGS); \
			}

#define	PROCWRAP_FUNC(FUNC,NUM) \
		FUNC ## tbl[NUM]

#if MAXBOARDS <= 4
	#define	PROCWRAP_FUNCTIONS(FUNC) \
			PROCWRAP_FUNCTION(FUNC, 0) \
			PROCWRAP_FUNCTION(FUNC, 1) \
			PROCWRAP_FUNCTION(FUNC, 2) \
			PROCWRAP_FUNCTION(FUNC, 3) \

	#define	PROCWRAP_TBL(FUNC) \
			static read_proc_t * FUNC ## tbl[MAXBOARDS] = \
			{ \
				FUNC ## 0, \
				FUNC ## 1, \
				FUNC ## 2, \
				FUNC ## 3, \
			};

	#define	PROCWRAPPERS(FUNC) \
			PROCWRAP_FUNCTIONS(FUNC) \
			PROCWRAP_TBL(FUNC)
#elif MAXBOARDS <= 8
	#define	PROCWRAP_FUNCTIONS(FUNC) \
			PROCWRAP_FUNCTION(FUNC, 0) \
			PROCWRAP_FUNCTION(FUNC, 1) \
			PROCWRAP_FUNCTION(FUNC, 2) \
			PROCWRAP_FUNCTION(FUNC, 3) \
			PROCWRAP_FUNCTION(FUNC, 4) \
			PROCWRAP_FUNCTION(FUNC, 5) \
			PROCWRAP_FUNCTION(FUNC, 6) \
			PROCWRAP_FUNCTION(FUNC, 7) \

	#define	PROCWRAP_TBL(FUNC) \
			static read_proc_t * FUNC ## tbl[MAXBOARDS] = \
			{ \
				FUNC ## 0, \
				FUNC ## 1, \
				FUNC ## 2, \
				FUNC ## 3, \
				FUNC ## 4, \
				FUNC ## 5, \
				FUNC ## 6, \
				FUNC ## 7, \
			};

	#define	PROCWRAPPERS(FUNC) \
			PROCWRAP_FUNCTIONS(FUNC) \
			PROCWRAP_TBL(FUNC)
#else
	#error "Fix proc.c to handle MAXBOARDS > 8"
#endif

/*
 *	In place conversion of spaces to tabs
 */
static int
spaces2tabs(char *buf)
{
	char	*ip, *op;
	int	col, tcol;

	ip = op = buf;
	col = 0;
	while (*ip)
	{
		if (*ip == ' ')
		{
			tcol = col;
			while (*ip == ' ')
			{
				++ip;
				++col;
				if ( (col&7) == 0)
				{
					if (col-tcol > 1)
						*op++ = '\t';
					else
						*op++ = ' ';
					tcol = col;
				}
			}
			while (tcol != col)
			{
				*op++ = ' ';
				++tcol;
			}
		}
		else if (*ip == '\t')
		{
			col = (col & ~7) + 8;
			*op++ = *ip++;
		}
		else
		{
			*op++ = *ip++;
			++col;
		}
	}
	*op = 0;
	return (op - buf);
}

/*
 * convert unsigned long long to strings for printing purposes
 */
STATIC void format_qu_uint64(char *buf, uint64 num)
{
        char    work[24];
        char    *workend = &work[sizeof(work) - 1];
        char    *w;
 
        /* Put the number in WORK. */
        w = workend;
        do {
                *w-- = (num % 10) + '0';
                num /= 10;
        } while (num > 0);
 
        /* Now write it to the caller, left justified */
        while (++w <= workend)
                *buf++ = *w;
        *buf = 0;
}



/*
 * /proc/dgdm/<boardno>/tty
 */
STATIC struct proc_dir_entry *ProcDgDmTTY[MAXBOARDS];

STATIC int dgdm_proc_tty(int brdno, PROC_READ_PROTO)
{
	BOARD_T	*brd;
	DRVCHAN	*dc;
	char	*p = buf;
	char	*e = buf + PAGE_SIZE - 3 * 40;
	char	*tp;
        char    conv[24];
	int	i, j;

	DPR_PROC(("dgdm_proc_tty count=%d fpos=%ld\n", count, fpos));

        brd = dgdm_Board[brdno];

        if(!brd)
                return(-EINVAL);

	/* Prepare the Header Labels */
	p += sprintf(p, "         Receive Statistics        "
				 "Transmit Statistics\n");
	p += sprintf(p, "%2s %19s %7s %19s %7s %s\n",
                     "Ch", "Chars Rx", "RxCPS",
                     "Chars Tx", "TxCPS", " Line Status Flags");
	
	for (dc = brd->dc, i=0; i < brd->nttychans; ++i, ++dc) {
		if (p > e) break;
		tp = p;
		p += sprintf(p, "%2d ", i);
		for (j = 0; j < NSTAT; ++j) {
			/* convert the long long into a string for printing */
			format_qu_uint64(conv, dc->stat[j].cnt);
			p += sprintf(p, "%19s %7ld ",
				conv, dc->stat[j].sma / (SMA_DUR/HZ) );
                }
		if(!dc->opencnt) {
			p += sprintf(p, " -- -- -- -- -- -- --") ;
		} else {
			p += sprintf(p, " op %s %s %s %s %s %s",
				     dc->mstat&dc->m_rts?"rs":"--",
				     dc->mstat&dc->m_cts?"cs":"--",
				     dc->mstat&dc->m_dtr?"tr":"--",
				     dc->mstat&dc->m_dsr?"mr":"--",
				     dc->mstat&dc->m_dcd?"cd":"--",
				     dc->mstat&dc->m_ri ?"ri":"--");
		}
		p += sprintf(p, "\n");
		/* Hack to compress output enough for 60 channels */
		p = tp + spaces2tabs(tp);
	}

	DPR_PROC(("dgdm_proc_tty returns %d\n", p-buf));

	return(p-buf);
}

PROCWRAPPERS(dgdm_proc_tty)


/*
 * /proc/dgdm/<boardno>/ttys
 */
STATIC struct proc_dir_entry *ProcDgDmTTYs[MAXBOARDS];

STATIC int dgdm_proc_ttys(int brdno, PROC_READ_PROTO)
{
	BOARD_T	*brd;
	DRVCHAN	*dc;
	char	*p = buf;
	char	*e = buf + PAGE_SIZE - 3 * 40;
	int	i;

	DPR_PROC(("dgdm_proc_ttys count=%d fpos=%ld\n", count, fpos));

        brd = dgdm_Board[brdno];

        if(!brd)
                return(-EINVAL);

	p += sprintf(p, "Chan\t&tty\tflags\n");

	for (dc = brd->dc, i=0; i < brd->nchans; ++i, ++dc) {
		if (p > e) break;
		p += sprintf(p, "%d\t0x%x\t0x%lx\n",
			i, (int) dc->tty,
			dc->tty ? dc->tty->flags : 0);
	}

	DPR_PROC(("dgdm_proc_ttys returns %d\n", p-buf));

	return(p-buf);
}

PROCWRAPPERS(dgdm_proc_ttys)

/*
 * /proc/dgdm/<boardno>/dxb
 */
STATIC struct proc_dir_entry *ProcDgDmDxb[MAXBOARDS];

STATIC int dgdm_proc_dxb(int brdno, PROC_READ_PROTO)
{
	BOARD_T	*brd;
	char	*p = buf;

	DPR_PROC(("dgdm_proc_dxb\n"));

        brd = dgdm_Board[brdno];

        if(!brd)
                return(-EINVAL);

	p += sprintf(p, "Deferred Polls:\t%d\n", brd->deferred_polls);
	p += sprintf(p, "Interrupts:\t%d\n", brd->intr_count);

	p += sprintf(p, "&Dxb:\t\t%x\n", (int) brd->dxb);
	p += sprintf(p, "Dxb->chan:\t%x\n", (int) brd->dxb->chan);

	/*
         * The in and out pointers in libdxb are uint32's  (and not 
	 * host-dependent sizes) so we should always be able to display
	 * them as 8 characters hex numbers.
	 */
	p += sprintf(p, "H2A:\t\tin=%8ux	out=%8ux\n",
                     brd->dxb->h2a.in, brd->dxb->h2a.out);
	p += sprintf(p, "A2H:\t\tin=%8ux	out=%8ux\n",
                     brd->dxb->a2h.in, brd->dxb->a2h.out);

	return(p-buf);
}

PROCWRAPPERS(dgdm_proc_dxb)


/*
 * /proc/dgdm/<boardno>/info
 *
 * Variables compatible with the shell
 */
STATIC struct proc_dir_entry *ProcDgDmBrdInfo[MAXBOARDS];

STATIC int dgdm_proc_brd_info(int brdno, PROC_READ_PROTO)
{
	BOARD_T		*brd;
	char		*p = buf;
	char		*name;

	DPR_PROC(("dgdm_proc_brd_info\n"));

        brd = dgdm_Board[brdno];

        if(!brd)
                return(-EINVAL);

	switch (brd->board_type) {
	case POST_TYPE_HOUSTON_T1:
	case POST_TYPE_HOUSTON_E1:
		name = DIGI_PCIHOUSTON_FAMILY_NAME;
		break;
    case POST_TYPE_CAMPINAS_T1:
    case POST_TYPE_CAMPINAS_E1:
        name = DIGI_CPCIHOUSTON_FAMILY_NAME;
        break;
	case POST_TYPE_LONDON_U:
	case POST_TYPE_LONDON_ST:
		name = DIGI_LONDON_FAMILY_NAME;
		break;
	case POST_TYPE_CHASKA:
		name = DIGI_CHASKA_FAMILY_NAME;
		break;
	case POST_TYPE_AP2K2_232:
	case POST_TYPE_AP2K4_232:
	case POST_TYPE_AP2K8_232:
	case POST_TYPE_AP2K16_232:
#if defined(POST_TYPE_AP2K4_MPI)	/* Temp until headers get caught up */
        case POST_TYPE_AP2K32_232:
        case POST_TYPE_AP2K4_MPI:
        case POST_TYPE_AP2K8_MPI:
        case POST_TYPE_AP2K16_MPI:
#endif
		name = DIGI_AP2K_FAMILY_NAME;
		break;
	default:
		name = "unknown";
		break;
	}
	p += sprintf(p, "DGDM_BOARDNAME=%s\n", name);
	p += sprintf(p, "DGDM_BOARDTYPE=%d\n", brd->board_type);

	/*
	 * report some things about the PCI bus that are important
	 * to some applications
	 */
        p += sprintf(p, "DGDM_VENDOR=0x%x\n", brd->vendor);
        p += sprintf(p, "DGDM_DEVICE=0x%x\n", brd->device);
        p += sprintf(p, "DGDM_SUBVENDOR=0x%x\n", brd->subvendor);
        p += sprintf(p, "DGDM_SUBDEVICE=0x%x\n", brd->subdevice);

	/*
	 * only report geographic address for the cPCI boards
	 */
	switch (brd->board_type) {
        case POST_TYPE_CAMPINAS_T1:
        case POST_TYPE_CAMPINAS_E1:
		p += sprintf(p, "DGDM_GEOADDR=0x%02x\n", brd->geoaddr);
		break;
	default:
		break;
	}

	/*
	 * report the physical addresses assigned to us when we got
	 * registered
	 */	
        p += sprintf(p, "DGDM_RAM=0x%lx\n", brd->ram);
        p += sprintf(p, "DGDM_REG=0x%lx\n", brd->reg);
        p += sprintf(p, "DGDM_FLASH=0x%lx\n", brd->flash);

	/*
	 * report available resources on the card
	 */
	p += sprintf(p, "DGDM_NDSPS=%d\n", brd->ndsps);
	p += sprintf(p, "DGDM_NCHANS=%d\n", brd->nttychans);

	return(p-buf);
}

PROCWRAPPERS(dgdm_proc_brd_info)


/*
 * /proc/dgdm/<boardno>/dsp
 */
STATIC struct proc_dir_entry *ProcDgDmDSP[MAXBOARDS];

STATIC int dgdm_proc_dsp(int brdno, PROC_READ_PROTO)
{
	BOARD_T		*brd;
	char		*p = buf;
	char		*e = buf + 4096 - 3 * 40;
	int		i;
	GLOBAL_COMMON	*gcp;
	int		daughter;
	uint32		reset[2];
	uint32		pwd[2];

	DPR_PROC(("dgdm_proc_dsp %d\n", count));

        brd = dgdm_Board[brdno];

        if(!brd)
                return(-EINVAL);

	gcp = dgdm_get_gc_ptr(brd);
	if (!gcp) return (-ENXIO);

	for (daughter = 0; daughter < 2; ++daughter) {
		if (p > e) break;

		gcp->rw.rw_type = RW_RAM;
		gcp->rw.rw_addr = 0xbfc80400 + daughter * 0x40000;
		gcp->rw.rw_size = 8;
		gcp->rw.rw_req = RW_READ;

		for (i = 0; gcp->rw.rw_req; ++i) {
			if (i == 10000) {
				if (dgdm_sleep(1))
					return (-EINTR);
				i = 0;
			}
		}

		memcpy(&reset[daughter], ((ulong *) gcp->rw.rw_data) + 0, 4);
		memcpy(&pwd[daughter], ((ulong *) gcp->rw.rw_data) + 1, 4);

		p += sprintf(p, "Bank %d\n", daughter);
		p += sprintf(p, "\tReset/ = %08ux\n", reset[daughter]);
		p += sprintf(p, "\tPWD/   = %08ux\n", pwd[daughter]);
	}

	return(p-buf);
}

PROCWRAPPERS(dgdm_proc_dsp)

/*
 * /proc/dgdm/<boardno>/dxbstrm
 */
STATIC struct proc_dir_entry *ProcDgDmDxbStrm[MAXBOARDS];

STATIC int dgdm_proc_dxbstrm(int brdno, PROC_READ_PROTO)
{
	BOARD_T	*brd;
	DXBSTRM	*ds;
	char	*p = buf;
	char	*e = buf + PAGE_SIZE - 3 * 40;
        char    conv[24];
	int	i, j;

	DPR_PROC(("dgdm_proc_dxbstrm count=%d fpos=%ld\n", count, fpos));

        brd = dgdm_Board[brdno];

        if(!brd)
                return(-EINVAL);

	/* Prepare the Header Labels */
	p += sprintf(p, "  "
			"     Receive Statistics........."
			"     Transmit Statistics........\n");
	p += sprintf(p, "%2s %9s %5s %9s %5s %9s %5s %9s %5s\n",
			"Ch",
			"Bytes", "BPS", "Pkts", "PPS",
			"Bytes", "BPS", "Pkts", "PPS");
	
	for (ds = brd->dxbisdn_chans, i = 0; i < brd->nttychans; ++i, ++ds) {
		if (p > e) break;
		p += sprintf(p, "%2d ", i);
		for (j = 0; j < NSTAT_Str; ++j) {
			/* convert the long long into a string for printing */
			format_qu_uint64(conv, ds->stat[j].cnt);
			p += sprintf(p, "%9s %5ld ",
                                     conv, ds->stat[j].sma / (SMA_DUR/HZ) );
                }
		p += sprintf(p, "\n");
	}

	DPR_PROC(("dgdm_proc_dxbstrm returns %d\n", p-buf));

	return(p-buf);
}

PROCWRAPPERS(dgdm_proc_dxbstrm)

/*
 * /proc/dgdm/info
 *
 *	Variables compatible with the shell
 */
STATIC struct proc_dir_entry *ProcDgDmInfo;

STATIC int dgdm_proc_info(PROC_READ_PROTO)
{
	char	*p = buf;

	DPR_PROC(("dgdm_proc_info\n"));
	p += sprintf(p, "MAXBOARDS=%d\n", MAXBOARDS);
	p += sprintf(p, "DGDM_NUMBOARDS=%d\n", dgdm_NumBoards);
	return(p-buf);
}

/*
 * /proc/dgdm/<boardno>/mknod
 */
STATIC struct proc_dir_entry *ProcDgDmBrdMknod[MAXBOARDS];

STATIC int mknod_cmd(char *buf, int yes, char *nodename, int boardnum,
                     int major, int firstminor, int count)
{
	if (!yes) return 0;

	return sprintf(buf, "dgdm_mknod\t%s\t%d\t%d\t%d\t%d\n",
		       nodename, boardnum, major, firstminor, count);
}

STATIC int ln_cmd(char *buf, int yes, char *src, char *dst, int  boardnum,
                  int portnum)
{
#define	LN_CMD_NONE	0x0
#define	LN_CMD_BOARD	0x1
#define	LN_CMD_PORT	0x2
#define	LN_CMD_BOTH	0x3
#define LN_SIZE		64

        int	prep = 0;
        int	order = 0;
        char	*p;
        char	tmp[LN_SIZE];
        char	scratch[LN_SIZE];
	char	srcname[LN_SIZE];
	char	dstname[LN_SIZE];


	if (!yes)
		return(0);

        /* prep source string */
        memset(scratch, '\0', LN_SIZE);
	memcpy(scratch, src, strlen(src));
        p = scratch;
        do {
                if(*p == '%') {
                        p++;
                        switch(*p) {
                        case 'b':
                                *p = 'd';
                                if(!order)
                                        order = LN_CMD_BOARD;
                                prep |= LN_CMD_BOARD;
                                break;
                        case 'p':
                                *p = 's';
                                if(!order)
                                        order = LN_CMD_PORT;
                                prep |= LN_CMD_PORT;
                                break;
                        default:
                                /* leave it alone if we're not sure */
                                prep |= LN_CMD_NONE;
                        }
                }
                p++;
        } while(*p != '\0');


        /* fill in the blanks */
        switch(prep) {
        case LN_CMD_BOARD:
                sprintf(srcname, scratch, boardnum);
                break;
        case LN_CMD_PORT:
                sprintf(tmp, scratch, "%02d");
                sprintf(srcname, tmp, portnum);
                break;
        case LN_CMD_BOTH:
                if(order == LN_CMD_BOARD)
                        sprintf(tmp, scratch, boardnum, "%02d");
                else
                        sprintf(tmp, scratch, "%02d", boardnum);
                sprintf(srcname, tmp, portnum);
                break;
 
        default:
                memcpy(srcname, src, strlen(src));
                break;
        }

	sprintf(dstname, dst, boardnum);
	return sprintf(buf, "ln -sf %s %s\n", srcname, dstname);
}

STATIC int dgdm_proc_brd_mknod(int brdno, PROC_READ_PROTO)
{
	BOARD_T	*brd = dgdm_Board[brdno];
        int	major;
	int	bn;
	char	*p = buf;

	DPR_PROC(("dgdm_proc_brd_mknod(%d) %d\n", brdno, count));

        if(!brd)
                return(-ENODEV);

        bn = brd->boardnum;

        /* 
         * if we're using one of the high 4 boards, temporarilly shift
         * the base major to build the nodes properly.
         */
        if (bn < 4)
                major = dgdm_major;
        else
                major = dgdm_major + DGDM_SERIAL2_MAJOR;

	/*
	 * Output the boilerplate at the top of the shell script
	 */
	p += sprintf(p, "#!/bin/sh\n\n");

	p += sprintf(p, "# /proc/dgdm/mknod [-d]\n");
	p += sprintf(p, "#\tMake dense modem device nodes.\n");
	p += sprintf(p, "#\t-d deletes all existing nodes first\n\n");

	p += sprintf(p, "PATH=" SBINDIR ":$PATH\n");

	p += sprintf(p, "[ -d %s ] || mkdir -p -m 755 %s\n",
                     DEVSTR, DEVSTR);

	p += sprintf(p, "[ \"$1\" != -d ] || rm -rf %s/* || exit 1\n",
                     DEVSTR);

	p += sprintf(p, "cd %s || exit 2\n\n", DEVSTR);

	/*
	 * For each board, output the device information in
	 * a handy table format...
	 */
	p += sprintf(p,	"#\t\tNAME\t\tBOARD\tMAJOR\tMINOR\tCOUNT\n");

        /* TTY devices and the boozer... */
        if(dgdm_tty) {
                p += mknod_cmd(p, dgdm_tty, dgdm_indevice, bn, 
                               (major + DGDM_SERIAL_MAJOR),
                               brd->firstminor, brd->nttychans);
                        
                p += mknod_cmd(p, dgdm_tty, dgdm_outdevice, bn,
                               (major + DGDM_CALLOUT_MAJOR),
                               brd->firstminor, brd->nttychans);
                        
                p += mknod_cmd(p, dgdm_tty, dgdm_specdevice, bn,
                               (major + DGDM_SERIAL_MAJOR),
                               (brd->firstminor + brd->nttychans),
                               brd->nspecchans);
                        
                /* Handy nicknames for the boozer */
                p += ln_cmd(p, dgdm_tty, dgdm_specdevice,
                            "beer%d", bn, 1);
                p += ln_cmd(p, dgdm_tty, dgdm_specdevice,
                            "lush%d", bn, 2);
                p += ln_cmd(p, dgdm_tty, dgdm_specdevice,
                            "vomit%d", bn, 3);

                if(dgdm_print) {
                        p += mknod_cmd(p, dgdm_tty, dgdm_printdevice, bn,
                                       (major + DGDM_PRINT_MAJOR),
                                       brd->firstminor, brd->nttychans);
                }
        }

        /* ISDN devices */
        if(dgdm_isdn) {
                p += mknod_cmd(p, dgdm_tty && dgdm_isdn,
                               dgdm_rawdxbdevice, bn,
                               (major + DGDM_ISDN_MAJOR),
                               board2minor(bn), NISDN);
        }

        /* Debugging/control/ram/flash devices ... */
        if (dgdm_control) {
                p += mknod_cmd(p, TRUE,  dgdm_ramdevice, bn, 
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn) | type2minor(TYPE_RAM)), 1);
                
                p += mknod_cmd(p, TRUE,  dgdm_dspramdevice, bn, 
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn) | 
                                type2minor(TYPE_RAM) | SUBTYPE_DSPDM), 2);

                p += mknod_cmd(p, TRUE,  dgdm_gdbdevice, bn, 
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn)	| 
                                type2minor(TYPE_RAM) | SUBTYPE_GDB), 1);

                p += mknod_cmd(p, TRUE,  dgdm_dmabufdevice, bn, 
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn) | type2minor(TYPE_DMABUF)), 2);

                p += mknod_cmd(p, TRUE,  dgdm_regdevice, bn, 
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn) | type2minor(TYPE_REG)),
                               SUBTYPE_NREG);

                p += mknod_cmd(p, TRUE,  dgdm_iodevice, bn,
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn) | type2minor(TYPE_IO)),
                               SUBTYPE_NIO);

                p += mknod_cmd(p, TRUE,  dgdm_flashdevice, bn, 
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn) | type2minor(TYPE_FLASH)),
                               SUBTYPE_NFLASH);

                p += mknod_cmd(p, brd->ndsps >= 24, dgdm_vpdalldevice, bn,
                               (major + DGDM_CONTROL_MAJOR),
                               (board2minor(bn) | type2minor(TYPE_VPD)),
                               VPD_AREAS);

                /* Handy nicknames for the special I/O devices */
                p += ln_cmd(p, TRUE, dgdm_iodevice, "print%d", bn, 1);
                p += ln_cmd(p, TRUE, dgdm_iodevice, "h2a%d", bn, 2);
                p += ln_cmd(p, TRUE, dgdm_iodevice, "a2h%d", bn, 3);
                p += ln_cmd(p, dgdm_tty, dgdm_iodevice, "audiotap%d", bn, 4);
                p += ln_cmd(p, dgdm_tty, dgdm_iodevice, "mgmt%d", bn, 5);
                p += ln_cmd(p, dgdm_tty, dgdm_iodevice, "eqm%d", bn, 6);

                /* first mgmt channel is used by dpa */
                if(bn == 0) {
                        p += ln_cmd(p, dgdm_tty, "mgmt%b", "dxbdl",
                                    bn, 5 /* not used */);
                }

                /* Handy nicknames for flash partitions */
                p += ln_cmd(p, TRUE, dgdm_flashdevice, "flash%d.all", bn, 0);
                p += ln_cmd(p, TRUE, dgdm_flashdevice, "flash%d.post", bn, 1);
                p += ln_cmd(p, TRUE, dgdm_flashdevice, "flash%d.parm", bn, 3);
                p += ln_cmd(p, TRUE, dgdm_flashdevice, "flash%d.boot", bn, 4);
                p += ln_cmd(p, TRUE, dgdm_flashdevice, "flash%d.devious",bn,5);
                p += ln_cmd(p, TRUE, dgdm_flashdevice, "flash%d.cfg", bn, 6);

                /* Handy nicknames for direct access to DSPs */
                p += ln_cmd(p, TRUE, dgdm_dspramdevice, "dsp%d.dm", bn, 0);
                p += ln_cmd(p, TRUE, dgdm_dspramdevice, "dsp%d.pm", bn, 1);

                /* Handy nickname for access to VPD flash */
                if (brd->ndsps >= 24)
                        p += ln_cmd(p, TRUE, dgdm_vpdalldevice, "vpd%d", bn,1);
                else
                        p += ln_cmd(p, TRUE, dgdm_flashdevice, "vpd%d", bn, 2);
        }

	return(p-buf);
}

PROCWRAPPERS(dgdm_proc_brd_mknod)


/*
 *      /proc/dgdm/mknod
 */
static struct proc_dir_entry *ProcDgDmMknod;

STATIC int dgdm_proc_mknod(PROC_READ_PROTO)
{
        char    *p = buf;

        DPR_PROC(("dgdm_proc_mknod\n"));

        p += sprintf(p, "#!/bin/sh\n\n");

        p += sprintf(p, "#\t/proc/dgdm/mknod [-d]\n");
        p += sprintf(p, "#\t\tMake dense modem device nodes.\n");
        p += sprintf(p, "#\t\t-d deletes all existing nodes first\n\n");
 
        p += sprintf(p, "[ \"$1\" != -d ] || rm -rf /dev/%s/* || exit 1\n",
                DRVSTR);

        p += sprintf(p, "for i in /proc/dgdm/*/mknod\n");
        p += sprintf(p, "do\n");
        p += sprintf(p, "       $i\n");
        p += sprintf(p, "done\n");

        return (p-buf);
}


/*
 * The directory entries /proc/dgdm/<boardnum>
 */
STATIC struct proc_dir_entry *ProcDgDmBrd[MAXBOARDS];

/*
 * The directory entry /proc/dgdm
 */
STATIC struct proc_dir_entry *ProcDgDm;

/*
 * Gee, the proc I/F for drivers is *still* not well done in 2.1.x.
 *
 * In any event, we'll *make* a better interface here and then
 * use it to abstract the differences between 2.0.x and 2.1.x
 *
 * Unfortunately, the *best* interface would also allow you
 * to pass in "private" data.  But that can't be supported
 * on 2.0.x, so that is why this interface is just "better".
 */
#if LINUX_VERSION_CODE >= 0x20100

STATIC struct proc_dir_entry *better_create_proc_entry(
        const char *name, mode_t mode, unsigned long size,
        read_proc_t *read_proc,	struct proc_dir_entry *parent)
{
        struct proc_dir_entry	*pde;

        pde = create_proc_entry(name, mode, parent);
        if (!pde) return NULL;

        if (size)
                pde->size = size;
        if (read_proc)
                pde->read_proc = read_proc;

        return pde;
}

STATIC void better_remove_proc_entry(struct proc_dir_entry *pde)
{
        if (!pde) return;

        remove_proc_entry(pde->name, pde->parent);
}

#else

STATIC struct proc_dir_entry *better_create_proc_entry(
        const char *name, mode_t mode, unsigned long size,
        read_proc_t *read_proc, struct proc_dir_entry *parent)
{
        struct proc_dir_entry	*pde;
        int			namelen;

        /* Allocate a proc_dir_entry and space for its name */
        namelen = strlen(name);
        pde = kmalloc(sizeof(struct proc_dir_entry) + namelen + 1, GFP_KERNEL);
        if (!pde)
                return NULL;

        /* Fill in the entry */
        memset(pde, 0, sizeof(struct proc_dir_entry));

        memcpy(((char *) pde) + sizeof(*pde), name, namelen + 1);
        pde->name = ((char *) pde) + sizeof(*pde);
        pde->namelen = namelen;

        pde->nlink = 1;
        if (mode & S_IFDIR) {
                mode |= S_IRUGO | S_IXUGO;
                /*
                 * NOTE: by rights, we should be using
                 *	 proc_dir_inode_operations for the ops.
                 *	 Sadly, this symbol is not exported in
                 *	 Linux 2.0.32.  proc_scsi_inode_operations
                 *	 seems to be a safe (if not entirely what
                 *	 we want) substitute that *is* exported.
                 */
                pde->ops = &proc_scsi_inode_operations;
                pde->nlink = 2;
        } else if (mode == 0) {
                mode = S_IFREG | S_IRUGO;
        }
        pde->mode = mode;

        if (size)
                pde->size = size;
        if (read_proc)
                pde->get_info = read_proc;

        if (parent == NULL)
                parent = &proc_root;

        proc_register_dynamic(parent, pde);

        return pde;
}

STATIC void better_remove_proc_entry(struct proc_dir_entry *pde)
{
        if (!pde) return;

        proc_unregister(pde->parent, pde->low_ino);
		kfree(pde);
}

#endif

/*
 * Register the basic /proc/dgdm files that appear whenever
 * the driver is loaded.
 */
GLOBAL void dgdm_proc_register_basic(void)
{
	char	*buf;
	int	size;
	int	i;
	
	/* Register /proc/dgdm */
	ProcDgDm = better_create_proc_entry(PROCSTR, S_IFDIR, 0, NULL, 0);

	/* Register /proc/dgdm/info */
	ProcDgDmInfo = better_create_proc_entry( "info", 0, 0,
			dgdm_proc_info, ProcDgDm);

	/*
	 * Compute size of /proc/dgdm/mknod output, fixup
	 * proc_dir_entry for it, then register it.
	 */
	buf = kmalloc(4096, GFP_KERNEL);
	if (buf) {
#if LINUX_VERSION_CODE >= 0x20100
                size = dgdm_proc_mknod(buf, NULL, 0, 4096, 0, 0);
#else
                size = dgdm_proc_mknod(buf, NULL, 0, 4096, 0);
#endif
	} else
		size = 0;
	kfree(buf);

	ProcDgDmMknod =
                better_create_proc_entry("mknod", S_IFREG | S_IRUGO | S_IXUGO,
                                         size, dgdm_proc_mknod, ProcDgDm);

	/*
	 * Register /proc/dgdm/<boardnum> directories
	 * and the /proc/dgdm/<boardnum>/info files
	 */
	for (i = 0; i < dgdm_NumBoards; ++i) {
		char	name[2];

		name[0] = i + '0';
		name[1] = 0;
		ProcDgDmBrd[i] = better_create_proc_entry(name, S_IFDIR, 0,
                                                          NULL, ProcDgDm);

		ProcDgDmBrdInfo[i] = better_create_proc_entry("info", 0, 0,
                        PROCWRAP_FUNC(dgdm_proc_brd_info, i),
                        ProcDgDmBrd[i]);

		/*
                 * Compute size of /proc/dgdm/<boardno>/mknod output, fixup
                 * proc_dir_entry for it, then register it.
                 */
                buf = kmalloc(4096, GFP_KERNEL);
                if (buf) {
#if LINUX_VERSION_CODE >= 0x20100
                        size = dgdm_proc_brd_mknod(i, buf, NULL, 0, 4096, 0,0);
#else
                        size = dgdm_proc_brd_mknod(i, buf, NULL, 0, 4096, 0);
#endif
                } else
                        size = 0;
                kfree(buf);
 
                ProcDgDmBrdMknod[i] = better_create_proc_entry("mknod",
                                        S_IFREG | S_IRUGO | S_IXUGO, size,
                                        PROCWRAP_FUNC(dgdm_proc_brd_mknod, i),
                                        ProcDgDmBrd[i]);
	}
}

/*
 * Register the proc devices that appear only if we are
 * loading up a fep.
 */
GLOBAL void dgdm_proc_register_fep(void)
{
	int	i;

	for (i = 0; i < dgdm_NumBoards; ++i) {
		ProcDgDmTTYs[i] = better_create_proc_entry( "ttys", 0, 0,
			PROCWRAP_FUNC(dgdm_proc_ttys, i),
			ProcDgDmBrd[i]);

		ProcDgDmTTY[i] = better_create_proc_entry( "tty", 0, 0,
			PROCWRAP_FUNC(dgdm_proc_tty, i),
			ProcDgDmBrd[i]);

		ProcDgDmDxb[i] = better_create_proc_entry( "dxb", 0, 0,
			PROCWRAP_FUNC(dgdm_proc_dxb, i),
			ProcDgDmBrd[i]);

		ProcDgDmDSP[i] = better_create_proc_entry( "dsp", 0, 0,
			PROCWRAP_FUNC(dgdm_proc_dsp, i),
			ProcDgDmBrd[i]);

		ProcDgDmDxbStrm[i] = better_create_proc_entry( "dxbstrm", 0, 0,
			PROCWRAP_FUNC(dgdm_proc_dxbstrm, i),
			ProcDgDmBrd[i]);
	}
}

/*
 * Unregister all of our /proc/dgdm/{*} devices
 */
GLOBAL void dgdm_proc_unregister_all(void)
{
	int	i;

#define NUKE_ONE(PDE) \
	if (PDE) { better_remove_proc_entry(PDE); PDE = NULL; } else

	for (i = 0; i < dgdm_NumBoards; ++i) {
		NUKE_ONE(ProcDgDmDxbStrm[i]);
		NUKE_ONE(ProcDgDmDSP[i]);
		NUKE_ONE(ProcDgDmDxb[i]);
		NUKE_ONE(ProcDgDmTTY[i]);
		NUKE_ONE(ProcDgDmTTYs[i]);
		NUKE_ONE(ProcDgDmBrdMknod[i]);
		NUKE_ONE(ProcDgDmBrdInfo[i]);
		NUKE_ONE(ProcDgDmBrd[i]);
	}

	NUKE_ONE(ProcDgDmMknod);
	NUKE_ONE(ProcDgDmInfo);

	NUKE_ONE(ProcDgDm);
}
