/* -*- mode: C++; c-basic-offset: 4; tab-width: 8 -*- */
/* ex: set ts=8 sw=4: */

/*
 Copyright 1999-2000 by Digi International (www.digi.com)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the 
 implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/**********************************************************************
 * $Header: /eng/cvs/triboot/tribl.h,v 1.34 2001/04/05 15:19:17 randall Exp $
 */

#ifndef _TRIBL_H_

#define _TRIBL_H_

#include "debugcomm.h"
#include "triglbl.h"
#include "dxbglbl.h"
#include "fepglbl.h"
#include "pci_ids.h"
#include "mipsheader.hh"

#define FLEN		24
#define NDEVS		32	/* number of supported boards */
#define NBARS		6	/* up to how many ?? */

#define REG_LEN		0x10000
#define RAM_LEN		(1024 * 1024)

#define	MEG		(1024 * 1024)
#define MEGABYTE(a)	((a) * MEG)

#define WAIT_TIME	60		/* seconds to wait for boot */


/*
 * These macros are meant to provide driver writers a means of
 * specifying how PCI memory accesses are to be made.  (This is needed
 * by Solaris and Windows NT and perhaps other OSes.)  If such
 * accesses can be done by simply dereferencing a virtual address,
 * nothing need be done and the macros below should suffice.
 *
 * Parameters:
 *   addr:  address
 *   value: value to be written to the address
 *   id:    registered triboot ID
 *   type:  type of memory (TRIBL_MEM, TRIBL_REGS, TRIBL_FLASH, ...)
 */
#ifndef TRIBL_PCI_READ8
#  define TRIBL_PCI_READ8(addr,id,type) \
	*((volatile uint8 *)(addr))
#  define TRIBL_PCI_WRITE16(addr,value,id,type) \
	*((volatile uint16 *)(addr))=(value)
#  define TRIBL_PCI_READ16(addr,id,type) \
	*((volatile uint16 *)(addr))
#  define TRIBL_PCI_WRITE32(addr,value,id,type) \
	*((volatile uint32 *)(addr))=(value)
#  define TRIBL_PCI_READ32(addr,id,type) \
	*((volatile uint32 *)(addr))
#  define TRIBL_MEMCPY_TO_PCI(dst,src,len,id,type) \
	memcpy(dst,src,len)
#  define TRIBL_MEMCPY_FROM_PCI(dst,src,len,id,type) \
	memcpy(dst,src,len)
#endif

/* This should be an integral type big enough to hold a pointer.
   uint32 should work fine for 32-bit systems.  On 64-bit systems
   (e.g., SPARCV9 systems), you'll have to use something else.
*/
#if defined(__sparcv9)
#	define tribl_uintptr_t	uintptr_t	
#elif defined(HPUX)
#       define tribl_uintptr_t uintptr_t
#else
#	define tribl_uintptr_t	uint32
#endif


/*
 * Each supported piece of hardware will potentially require a different
 * procedure or possibly order in the delivery of its code.  The following
 * enum can be expanded.  Hint: BOOT, POST, etc are just mnemonics
 */
enum
{
    DL_START = -1,
    DL_BOOT,
    DL_POST,
    DL_FEP,
    DL_DONE,
    DL_POSTWAIT
};

#ifdef NEED_NAMES
/*
 * Mapping of binaries to target board types.
 */

static char *hlc_binary_names[] =
{
    "hlcb.bin",
    "hlcp.bin",
    "hlcf.bin",
    NULL
};

static char *rap2_binary_names[] =
{
    "tiggerb.bin",
    "tiggerp.bin",
    "tiggerf.bin",
    NULL
};

/*DBUstart: even though as far as I can see, these are NEVER passed upstream
	    anyway, may as well nominally comply with the silliness.  The REAL
	    names are #defined in dgdxb/adapinit. 
*/
static char *s2k_binary_names[] =
{
    "s2kb.bin",	/*DL_BOOT*/
    "s2kp.bin", /*DL_POST*/
    "s2kf.bin",	/*DL_FEP */
    NULL
};
/*DBUend*/
#endif

typedef struct
{
    GLOBAL_POST_DATA	pst;
    POST_ERROR_REPORT	err[POST_MAX_ENTRIES];
} TRIBL_POST;

typedef struct
{
    /*
     * this is a unique ID which is used to differentiate different
     * boards in the system.
     * It may be an address, a register, any integer, just so it
     * is unique.
     */

    uint32	id;

    /*
     * The board specific portion needed in the boot process.
     * NOTE: if a driver is really confident of its device, and if
     *       everything was in a PROM or flash, it is conceivable that
     *       none of this would be done.
     */

#   define TRIBL_REGISTER	1
#   define TRIBL_MAP_ADDRS	2
#   define TRIBL_UNMAP_ADDRS	3
#   define TRIBL_MADDRS_DONE	4
#   define TRIBL_FW_BOOT	5
#   define TRIBL_FW_POST	6
#   define TRIBL_FW_FEP		7
#   define TRIBL_DRIVER_UNLOAD	8
#   define TRIBL_PANIC		9
#   define TRIBL_WAIT		10
#   define TRIBL_WAIT_DONE	11
#   define TRIBL_DXBINIT	12
#   define TRIBL_BOARD_STRUCT	13
#   define TRIBL_POST_DATA	14
#   define TRIBL_UNREGISTER	15

    /* be certain to check status on receipt of this one !! */
#   define TRIBL_DONE		0xfffffffe

    uint32	state;

    /*
     * This will potentially hold PCI-ID and vendor and/or revision
     */

#   define PCI_DIGI	0x114f
    uint16	pci_vendor;

    /*
     * TIGGER8 is now a misnomer, it would be more accurate to say
     * it is a generic HERCULES product.
     *
     * Since I do not know who might be using this, I will leave it alone.
     * Suffice it to say that there exists a sub-id which differentiates the
     * different products with different port count, the information is
     * made available to the drivers after POST as 'nchan'.
     */


	#ifndef DIGI_AP2K_DID 
	# define DIGI_AP2K_DID 0x0040
	#endif

#   define PCI_TIGGER8	DIGI_AP2K_DID
    uint16	pci_device;

    /*
     * fep_code.1, fep_code.2, etc
     * this is the sequence sent by the driver, acked by the library
     *
     * This is seen as a monotonically increasing value.  If there is a
     * gap in sequence, or sequence-acks it is an unrecoverable error,
     * start all over again.
     */

    uint32	sequence;

    /*
     * This is  used by TRIBL_WAIT so the library can tell the
     * driver how long it needs to wait before being called again.
     */

    uint32	count;

    /*
     * File name needed by the library - hardware specific
     */

    uint8	file_name[FLEN];

#   define TRIBL_OK			0
#   define TRIBL_ERROR			1

#   define TRIBL_CODE_ERROR		2
    /* Incorrect version for this hardware */
#	define TRIBL_VER_ERROR		(TRIBL_CODE_ERROR | (1 << 8))
    /* CRC failed after download */
#	define TRIBL_CRC_ERROR		(TRIBL_CODE_ERROR | (2 << 8))
#	define TRIBL_CODE_TOO_BIG	(TRIBL_CODE_ERROR | (3 << 8))
#	define TRIBL_HEADER_ERROR	(TRIBL_CODE_ERROR | (4 << 8))

#   define TRIBL_EARLY_ERROR		3
#	define TRIBL_MEMORY_ERROR	(TRIBL_EARLY_ERROR | (1 << 8))
#	define TRIBL_PORT_ERROR		(TRIBL_EARLY_ERROR | (2 << 8))
#	define TRIBL_MEMORY_SIZE_ERROR	(TRIBL_EARLY_ERROR | (3 << 8))

#   define TRIBL_PROTO_ERROR		4
    /* A necessary BAR is missing */
#	define TRIBL_BAR_ERROR		(TRIBL_PROTO_ERROR | (1 << 8))
    /* Ran out of devices */
#	define TRIBL_NO_DEV		(TRIBL_PROTO_ERROR | (2 << 8))
    /* Attempt to use an unregistered ID */
#	define TRIBL_NOT_REGISTERED	(TRIBL_PROTO_ERROR | (3 << 8))
    /* Attempt to register a dupicate ID */
#	define TRIBL_DUPLICATE		(TRIBL_PROTO_ERROR | (4 << 8))
    /* No virtual addresses provided to TRIBL_MADDRS_DONE */
#	define TRIBL_NO_VADDR		(TRIBL_PROTO_ERROR | (5 << 8))
    /* A packet is received out of order */
#	define TRIBL_SEQUENCE_ERROR	(TRIBL_PROTO_ERROR | (6 << 8))

#   define TRIBL_UNSUPPORTED_HW		5
    /* PCI Vendor ID not supported */
#	define TRIBL_VENDOR_ERROR	(TRIBL_UNSUPPORTED_HW | (1 << 8))
    /* PCI Device ID  not supported */
#	define TRIBL_DEVICE_ERROR	(TRIBL_UNSUPPORTED_HW | (2 << 8))

#   define TRIBL_DXBINIT_ERROR		15
#   define TRIBL_BOOT_ERROR		16	/* bad magic or status */
#   define TRIBL_POST_ERROR		17	/* bad magic or status */
#   define TRIBL_FEP_ERROR		18	/* bad magic or status */

    uint32	status;

    /*
     * The number of bytes included in this transfer
     */

    uint32	size;

    /*
     * this is used by the interrupt service routine to provide the
     * caller with information on which device had an interrupt cleared.
     */

    uint32	int_index;

    /*
     * A pointer to the data referenced by size
     */

    void	*pdata;

    /*
     * keep a pointer to the call-back function
     * define the support functions needed.
     */

/*DBUstart: silence compiler */
#if defined(DELAY)
#undef DELAY
#endif
/*DBUend*/

#   define DELAY	0	/* the 2nd parameter is # of usecs */
#   define MAPP2V	1	/* phys addr - size */
#   define UNMAP	2	/* virt addr - size */

    void	(*tri_cb)(uint32, tribl_uintptr_t*, uint32*, uint32*, uint32*);
} TRIBL_BOARD;

typedef struct
{
    char	**fepname;
    char	**postname;
} BOARD_BIN_NAMES;

typedef struct brd
{
    /* This is the same ID as in TRIBL_BOARD. */
    uint32	id;
    uint16	device;
    uint16	padding;

    /*
     * the following fields, at least through to irq need to be
     * abstracted out of this structure in order to support a
     * wider audience.
     */
    uint32		ram;		/* BAR1 - physical RAM address */
    uint32		ramlen;		/* RAM length */
    void		*vram;		/* Virtual RAM address */

    uint32		reg;		/* BAR2 - register addresses */
    uint32		reglen;		/* Registers length */
    void		*vreg;		/* Virtual registers address */

    uint32		flash;		/* Physical flash address */
    uint32		flashlen;	/* Flash length in bytes */
    uint32		*vflash;	/* Virtual Flash address */

    uint32		rom;
    uint32		romlen;
    void		*vrom;

    uint32		b_0;		/* BAR5 placeholder */
    uint32		b_0len;
    void		*vb_0;

    uint32		b_1;		/* BAR6 placeholder */
    uint32		b_1len;
    void		*vb_1;

    uint32		vpd;		/* Physical VPD or eprom address */
    uint32		vpdlen;		/* VPD or eprom length in bytes */
    uint8		*vvpd;		/* Virtual VPD or eprom address */
    int			vpd_refcnt;	/* Usage count - mapped on open */
    void		*vpdarea;	/* VPD Flash information */

    int			irq;		/* Interrupt request number */
    int			intr_count;	/* Count of interrupts */

    /*
     *	Global common area in shared memory
     */

    uint32		gco;		/* Offset of global common on board */
    GLOBAL_COMMON	*gcp;		/* Ptr global common data */

    uint32		bigendian_board;/* Board is running big-endian */
    uint32		need_swap;
    uint32		ramtest_len;
    uint32		wait_run;	/* time to wait for running code */

    /*
     *	Global debug data in shared memory
     */

    GLOBAL_DEBUG_DATA	*gdd;		/* Ptr to area in shared memory */
    uint32		dma_buf[2];	/* Phyaddr of two 4k DMA buffers */

    int			nchans;		/* number of ports on board */

/*DBUstart*/
    uint32		gho;		/* Offset of global HostData on board */
    uint32		fco;		/* Offset  for loading FEP */
/*DBUend*/
    GLOBAL_HOST_DATA	*ghd;		/* Global host data */

    GLOBAL_FEP_DATA	*gfd;		/* Board global struct */

    /*
     *	Per-board function dispatches.
     */

    /* Post intr to board */
    void		(*post_board_int)(struct brd *);

    /* Clear pending intr from board */
    int			(*clear_board_int)(struct brd *);

    /* Unmask interrupts from host */
    void		(*unmask_board_int)(struct brd *);

    /* Mask interrupts from host */
    void		(*mask_board_int)(struct brd *);

    /* Put hardware in a halted state */
    void		(*shutdown_board)(struct brd *);

    /* reset and load address 0 with initial jump addr */
    void		(*start_board)(struct brd *, uint32);

    /* Actions taken after registers mapped */
    int			(*postmap)(struct brd *);

    /* Actions taken after memory mapped */
    int			(*postrammap)(struct brd *);

    /* further configure the hardware as necessary */
    int			(*configure)(struct brd *, uint32 a[]);

    /* a hook to put the board in reset */
    void		(*reset_board)(struct brd *, uint32 force);

    /* clear the reset */
    void		(*release_board_reset)(struct brd *);

    /* Pointer to Global Common Data */
    void *		(*get_gcd)(struct brd *);

    /* Pointer to Global Post Data */
    void *		(*get_gpd)(struct brd *);

    /* Pointer to Global FEP Data */
    void *		(*get_gfd)(struct brd *);

    /* The list of binaries for this board */
    BOARD_BIN_NAMES	bin_names;

    /* Which binary to start first on this board */
    int			dl_start;

    /* the callback function */
    void		(*cb)(uint32, tribl_uintptr_t*, uint32*, uint32*, uint32*);

    TRIBL_POST		*post;
} BOARD;

void tri_boot(TRIBL_BOARD *);
uint32 tribl_intr(uint32);

/*
 * A structure for the library to pass PCI addresses to the driver.
 * This will enable the library to decide which addresses are of what
 * type, and if the driver really needs to know, it may use them.
 *
 * The names other than MEMORY can probably be used in an arbitrary
 * fashion, the host side probably has no right to be using any of the
 * others without going through the library, once we are up and running.
 */

typedef struct
{
    tribl_uintptr_t  addr;	/* address of this register */
    uint32	size;		/* size of address space */

#   define TRIBL_RAWB	0	/* type not known, passing the value, */
				/* the address at BAR1,BAR2,etc */
#   define TRIBL_MEM	1	/* Dual ported */
#   define TRIBL_REGS	2	/* Maybe something close */
#   define TRIBL_FLASH	3
#   define TRIBL_ROM	4	/* E^2 or PROM, something similar */
#   define TRIBL_B_0	5	/* use these as you will */
#   define TRIBL_B_1	6
    uint32	type;
    uint32	mtag;		/* Tom wants it ?? */
} PADDRS;

struct d_info
{
    uint32	u_id;
    uint16	pci_vendor;
    uint16	pci_device;
#   define DXBSENT	1
#   define DXBDONE	2
#   define POSTMAP	4
    uint32	flags;
    uint32	cur_status;
    uint32	cur_state;
    uint32	cur_sequence;
    uint32	code;
    uint32	hp;		/* Current index into hdr */
    uint32	cur_down_addr;
    uint32	wait_count;
    uint32	intr_index;
    PADDRS	*paddrs;
    struct digiheader	hdr;
};

/*
 * There are values on the board(s) that are always little endian,
 * if the host is big endian these must be swapped when accessed.
 */

#ifdef sparc
#	define BIG_ENDIAN_HOST
#endif

#ifdef BIG_ENDIAN_HOST
#   define HOST_ENDIAN	GLOBAL_ENDIAN_BIG
#else
#   define HOST_ENDIAN	GLOBAL_ENDIAN_LITTLE
#endif

#ifndef SOLARIS
/* This is a bit troublesome as tribl.h gets #included into driver
 * files which may use DEBUG in a slightly different way (using #ifdef
 * instead of #if).  (It's also unnecessary as "#if UNDEFINED_SYMBOL"
 * gets expanded to "#if 0".)  Anyhow, this gums up the works in the
 * Solaris driver, so I'll simply exclude them.
 */

/* I commented this code out and changed the use of "DEBUG" in this header
 * file. "if DEBUG" has been changed to "ifdef DEBUG" below. Defining DEBUG
 * if it was not defined was giving the NOVELL driver grief and I suspect
 * the Solaris driver as well. Jeff McCracken 2/1/2000
 */

//#ifndef DEBUG
//#define DEBUG	0
//#endif

#endif /* #ifndef SOLARIS */

/*
 * Ideally, these should be from a makefile in the form -Dxxx
 * DEBUG should probably be the same
 */

// #define LINUX
// #define NOVELL
// #define UW7
// #define SCOUNIX
// #define LINUX
// #define SOLARIS

//#if DEBUG JPM 2/1/2000
#ifdef DEBUG
#   ifdef SCOUNIX
	extern int printf();
#	define myprintf	printf
#   endif
#   ifdef LINUX
#     ifdef DGDM_TRACER	
	extern void dgdm_tracef(const char *fmt, ...);
#	define myprintf	dgdm_tracef
#     else
	extern int printk();
#	define myprintf	printk
#     endif
#   endif
#   ifdef NOVELL
	extern ConsolePrintf();
#	define myprintf	ConsolePrintf
#   endif
#   ifdef UW7
	extern void dxb_printf();
#	define myprintf dxb_printf
#   endif
#   ifdef _AIX
	extern int uprintf();
#	define myprintf uprintf
#   endif
#   ifdef WIN32
#	define myprintf DbgPrint
#   endif
#   ifdef SOLARIS
#	define myprintf
#   endif
#   ifdef HPUX
#	define myprintf msg_printf
#   endif
#endif

#if defined(SCOUNIX) || defined(UW7)
extern void bcopy(const void *, void *, uint);
#define memcpy(a, b, c)		bcopy((b), (a), (c))
#if !defined(UW7)
	extern void bzero(void *, uint);
#	define memset(a,  b, c)	bzero((a), (c))
#	endif

#elif defined(SOLARIS)
#	ifndef memcpy
#		define memcpy(a, b, c)	bcopy((b), (a), (c))
#	endif
#	ifndef memset
#		define memset(a, b, c)	bzero((a), (c))
#	endif

#elif (!defined(LINUX) && !defined(HPUX))
extern void *memcpy(void *, const void *, uint);
extern void *memset(void *, int, uint);
#endif

//#if DEBUG JPM 2/1/2000
#ifdef DEBUG
#define dump_tb(tb) \
{ \
    myprintf("file(%s) line(%d)\n", __FILE__, __LINE__);	\
    myprintf("\tid		(%x)\n", (tb)->id);		\
    myprintf("\tstate	(%x)\n", (tb)->state);			\
    myprintf("\tvendor	(%x)\n", (tb)->pci_vendor);		\
    myprintf("\tdevice	(%x)\n", (tb)->pci_device);		\
    myprintf("\tsequence	(%x)\n", (tb)->sequence);	\
    myprintf("\tcount	(%x)\n", (tb)->count);			\
    myprintf("\tstatus	(%x)\n", (tb)->status);			\
    myprintf("\tsize	(%x)\n", (tb)->size);			\
    myprintf("\n\n");						\
}
#else
#define dump_tb(x)
#endif

#endif /* #ifndef _TRIBL_H_ */


/* #ifdef DEBUG */
/* #include <sys/cmn_err.h> */
/* #endif */
